import net.sf.saxon.s9api.*;
import java.io.File;
import java.io.IOException;
import java.io.BufferedReader;
import java.io.InputStreamReader;

public class XPathClient {
    /**
    * Read text from the standard input. Individual commands are terminated
    * by an empty line.
    *
    * @return The command read from standard input
    */
    private static String readCommand() throws IOException {
        BufferedReader inread = new BufferedReader(new InputStreamReader(System.in));
        String line = "";
        StringBuilder xpexpr = new StringBuilder();

        do {
            line = inread.readLine();

            if (line.length() != 0) {
                xpexpr.append("\n");
                xpexpr.append(line);
            }
        }
        while (line.length() != 0);

        if (xpexpr.length() > 1) {
            return xpexpr.toString().substring(1);
        }
        return "";
    }

    /**
    * Get a saxon node representation of a document, from a File.
    *
    * @param docfile The input file, containing XML
    * @param proc The processor used to create the document node.
    *
    * @return A node representing the document.
    */
    private static XdmNode loadDocument(File docfile, Processor proc) 
        throws SaxonApiException {

        DocumentBuilder builder = proc.newDocumentBuilder();
        builder.setLineNumbering(true);
        
        return builder.build(docfile);
    }

    public static void main( String args[] ) throws IOException {
        if (args.length != 1) {
            System.out.println("Utilisation: java XPathClient [document XML]");
            return;
        }

        Processor proc = new Processor(false);
        String xpathexpr;
        XPathCompiler xpath = proc.newXPathCompiler();
        xpath.setAllowUndeclaredVariables(true);
        XdmNode doc;
        String[] cmdargs;

        /* load the document */
        try {
            doc = loadDocument(new File(args[0]), proc);
        }
        catch(SaxonApiException e) {
            System.out.println(e.getMessage());
            return;
        }

        /* take XPath as input, until a blank line is entered */
        do {
            xpathexpr = readCommand();

            if (xpathexpr.startsWith("#")) {
                /* Some custom command occured... handle it */
                cmdargs = xpathexpr.split(" ");

                if (cmdargs[0].equals("#bind-ns") && (cmdargs.length == 3)) {
                    xpath.declareNamespace(cmdargs[1], cmdargs[2]);
                }
                else if (cmdargs[0].equals("#unbind-ns") && (cmdargs.length == 2)) {
                    xpath.declareNamespace(cmdargs[1], "");
                }
            }
            else if (xpathexpr.length() != 0) {
                try {
                    XPathSelector selector = xpath.compile(xpathexpr.toString()).load();
                    int i = 0;
                    selector.setContextItem(doc);

                    for (XdmItem item: selector) {
                    StringBuilder sb = new StringBuilder();

                    sb.append(++i + ": ");

                    if (item.isAtomicValue()) {
                        XdmAtomicValue xav = (XdmAtomicValue)item;
                        sb.append(xav.toString());
                        sb.append(" (");
                        sb.append(xav.getPrimitiveTypeName());
                        sb.append(")");
                    }
                    else {
                        XdmNode node = (XdmNode)item;
                        String nodeString = node.toString();
                        if (node.getNodeKind() == XdmNodeKind.ELEMENT) {
                            sb.append(node.getNodeName());
                            sb.append("\n");
                            if (nodeString.length() < 120) {
                                sb.append(nodeString);
                            }
                            else {
                                sb.append(nodeString.substring(0,120));
                                sb.append("...");
                            }
                        }
                        else {
                            sb.append(nodeString.replace("\n", ""));
                        }
                    }   
                    System.out.println(sb.toString());
                    }
                }
                catch (Exception e) {
                    System.out.println(e.getMessage());
                }
            }
            System.out.println();
        }
        while (xpathexpr.length() != 0);
    }
}
